#include <stdio.h>
#include <stdlib.h>		/* for malloc(3) */
#include <err.h>
#include "ac.h"

/* 	$Id: ac_print.c,v 1.2 2011/02/17 19:12:37 rec Exp $	 */

#ifndef lint
static char vcid[] = "$Id: ac_print.c,v 1.2 2011/02/17 19:12:37 rec Exp $";
#endif /* lint */


struct node {
  AC_priv_tree addr;
  unsigned num;
};

struct list {
  struct node *p;
  unsigned len;
};

static add_addr(struct list *l, AC_priv_tree addr, unsigned num)
{
  int m, i;
  struct node *p;

  if (l->p == NULL) {
    l->p = malloc(sizeof(struct node));
    l->len = 1;
    l->p->addr = addr;
    l->p->num = l->len;
    return;
  }

  p = l->p;
  for (i = 0; i < l->len; i++,p++)
    if (p->addr == addr) {
      printf("Duplicate insert.\n");
      return;
    }

  m = l->len;
  l->len += 1;
  l->p = realloc(l->p,l->len * sizeof(struct node));
  l->p[m].addr = addr;
  l->p[m].num = l->len;
}

static unsigned get_addr(struct list *l, AC_priv_tree addr)
{
  unsigned u, i;
  struct node *p;

  p = l->p;
  for (i = 0; i < l->len; i++,p++)
    if (p->addr == addr)
      return p->num;
  add_addr(l, addr, 0);
  return get_addr(l,addr);
}  

static void code_walk(struct list *l, 
		      AC_priv_tree p,
		      unsigned *node)
{
  if (p == NULL) return;
  add_addr(l, p->sibling, *node);
  code_walk(l,p->sibling, node);
  *node += 1;		/* get next node number */
  add_addr(l, p->children, *node);
  code_walk(l,p->children, node);
  *node += 1;		/* get next node number */
}

static void code_child(FILE *f, 
			  struct list *l,
			  char *var_name,
			  AC_priv_tree p,
			  unsigned n1, 
			  unsigned *node)
{
  unsigned n2;

  if (p == NULL) 		/* if no node here, end recursion */
    return;

  n2 = *node += 1;		/* get next node number */

  if (p->matchid)		/* is this a terminal? If so, mark it */
    fprintf(f,"/*\n * Terminal for %u\n */\n", n2);

  fprintf(f,"struct ac_priv_tree_node %s_%u = {\n", var_name, get_addr(l,p));
  fprintf(f,"    .outlink = &%s_%u,\n", var_name, get_addr(l, p->outlink));
  fprintf(f,"    .faillink = &%s_%u,\n", var_name, get_addr(l, p->faillink));
  fprintf(f,"    .children = &%s_%u,\n", var_name, get_addr(l, p->children));
  fprintf(f,"    .sibling = &%s_%u,\n", var_name, get_addr(l, p->sibling));
  fprintf(f,"};\n");

  code_child(f,l,var_name,p->children, n2, node); /* print remaining kids */
  code_child(f,l,var_name,p->sibling, n1, node);	/* then go for siblings */
}

void ac_code(AC_Handle p, 
	     char *var_name,	/* root for unique variables */
	     char *fname)
{
  FILE *f;			/* DOT format output file */
  unsigned node = 0;		/* incremented as nodes are processed */
  AC_priv_tree q;		/* pointer for traversal */
  int i;			/* loop variable */
  struct list list;		/* list of addresses */

  list.p = NULL;
  list.len = 0;
  code_walk(&list, p->tree->children, &node);
  printf("Added %d nodes.\n", list.len);
  node = 0;
  /*
   * Open output file.
   * Note: If cannot open output file, print error and return.
   * Failure to create graph should not crash an application.
   */

  if ((f = fopen(fname,"w")) == NULL) {
    warn("Unable to create output file %s for compiled C output.", fname);
    return;
  }
  /*
   * Print standard preamble
   */
  fprintf(f,"#include <stdio.h>        /* for our definitions */\n");
  fprintf(f,"#include <ac.h>        /* for our definitions */\n");
  fprintf(f,"\n");
  /*
   * Forward declarations
   */
  fprintf(f,"extern struct ac_priv_tree_node %s_%u", var_name, list.p[0].num);
  for (i = 1; i < list.len; i++) {
    fprintf(f,",%s_%u", var_name, list.p[i].num);
    if (i % 8 == 0) fprintf(f,"\n");
  }
  fprintf(f,";\n");

  /*
   * Print rest of state machine by recursion
   */
  code_child(f,&list, var_name,p->tree->children,0,&node);
  if (p->Psize > 0) {
    fprintf(f,"unsigned %s_lengths[] = {%u", var_name, p->Plengths[0]);
    for (i = 1; i < p->Psize; i++)
      fprintf(f,",%u", p->Plengths[i]);
    fprintf(f,"};\n");
  }
  /*
   * print root node.
   */
  fprintf(f,"AC_Struct %s = {         /* Accessed as an AC_Handle by pointer */\n",var_name);
  fprintf(f,"        .T = %#x,\n", p->T);
  fprintf(f,"        .N = %u,\n", p->N);
  fprintf(f,"        .Psize = %u,\n", p->Psize);
  if (p->Psize == 0) {
    fprintf(f,"        .Plengths = 0,\n");
  } else {
    fprintf(f,"        .Plengths = &%s_lengths,\n", var_name);
  }
  if (p->tree)
    fprintf(f,"        .tree = &%s_%u,\n", var_name, get_addr(&list,p->tree));
  else 
    fprintf(f,"        .tree = NULL,\n");
  if (p->w)
    fprintf(f,"        .w = &%s_%u,\n", var_name, get_addr(&list,p->w));
  else
    fprintf(f,"        .w = NULL,\n");
  if (p->output)
    fprintf(f,"        .output = &%s_%u,\n", var_name, get_addr(&list,p->output));
  else
    fprintf(f,"        .output = NULL,\n");
  fprintf(f,"};\n");
  /*
   * Terminate file and close
   */
  fclose(f);
  free(list.p);
}
